;;; gobject-cffi.lisp --- Emit gobject cffi bindings.
;;
;; Copyright (C) 2010 Jan Moringen
;;
;; Author: Jan Moringen <jmoringe@techfak.uni-bielefeld.de>
;;
;; This Program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.
;;
;; This Program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program. If not, see <http://www.gnu.org/licenses>.

(in-package :gir.emitter)

(defun eval-and-print (forms)
  (format t "evaluating ~S~%" forms)
  (eval forms))


;;; gobject + cffi
;;

(defmethod emit :before ((node namespace) (target (eql :gobject-cffi)))
  "Load the library corresponding to the namespace."
  (bind (((:accessors (shared-library namespace-shared-library)) node))
    (format t "loading library ~S~%" shared-library)
    (load-foreign-library shared-library)))

(defmethod emit ((node namespace) (target (eql :gobject-cffi)))
  (with-emitter-symbols
    (format t "switching to package ~S~%" package)
    (let ((*package*                   package)
	  (gobject:*lisp-name-package* package))
      (format t "now in package ~S~%" *package*)
      (call-next-method))))

(defmethod emit ((node alias) (target (eql :gobject-cffi)))
  (warn "emit is not implemented for alias nodes; ignoring ~A" node))

(defmethod emit ((node function1) (target (eql :gobject-cffi)))
  (with-emitter-symbols
    (bind (((:accessors (name         node-name)
			(foreign-name node-foreign-name)
			(return-type  function-return-type)
			(parameters   function-parameters)
			(doc          node-doc)) node)
	   (lisp-name       (gobject::g-name->name name))
	   (ffi-return-type (gir-type->ffi-type return-type))
	   (ffi-parameters  (mapcar #'recur parameters)))

      ;; TODO append argument documentation to docstring
      (eval-and-print
       `(defcfun (,lisp-name ,foreign-name) ,ffi-return-type
	  ,@(when doc `(,doc))
	  ,@ffi-parameters)))))

(defmethod emit ((node parameter) (target (eql :gobject-cffi)))
  (bind (((:accessors (name node-name)
		      (type parameter-type)) node)
	 (lisp-name (gobject::g-name->name name))
	 (ffi-type  (gir-type->ffi-type type)))
    `(,lisp-name ,ffi-type)))

(defmethod emit ((node type1) (target (eql :gobject-cffi)))
  (values))

#| Example:
 (define-g-enum "GtkTextDirection"
    text-direction
    (:export t :type-initializer "gtk_text_direction_get_type")
  (:none 0)
  (:ltr 1)
  (:rtl 2))
|#

(defmethod emit ((node enumeration) (target (eql :gobject-cffi)))
  (with-emitter-symbols
      (bind (((:accessors (name        node-name)
			(c-name      class-type-name)
			(type-getter class-type-getter)
			(children    composite-children)) node)
	   (lisp-name (gobject::g-name->name name))
	   (values (mapcar #'recur children)))

      (eval-and-print
       `(gobject:define-g-enum ,c-name ,lisp-name
	    (:export t ,@(when type-getter `(:type-initializer ,type-getter)))
	  ,@values)))))

(defmethod emit ((node member1) (target (eql :gobject-cffi)))
  (bind (((:accessors (name  node-name)
		      (nick  value-nick)
		      (value node-value)) node)
	 (lisp-name (intern (string-upcase (or nick name)) :keyword))) ;; TODO proper lispy mame
    `(,lisp-name ,(read-from-string value)))) ;; TODO do this when parsing

(defmethod emit ((node class1) (target (eql :gobject-cffi)))
  (with-emitter-symbols
    (bind (((:accessors
	     (type            node-name)
	     (foreign-name    node-foreign-name)
	     (type-name       class-type-name)
	     (type-getter     class-type-getter)
	     (impl-interfaces class-implemented-interfaces)
	     (parent          class-parent)
	     (properties      class-properties)
	     (fields          class-fields)) node)
	   (lisp-type (intern (string-upcase type) package))
	   ;;(properties (find-node)

	   (slot-specs (mapcar #'recur (append properties fields))))
      (eval-and-print
       `(gobject:define-g-object-class ,type-name ,lisp-type
	  (,@(when parent
		   `(:superclass ,parent))
	   :export           t
	   :interfaces       ,impl-interfaces
	   :type-initializer ,type-getter)
	  ,slot-specs)))))

;; Example
;; (define-g-object-class "GtkExpander" expander
;;   (:superclass bin :export t :interfaces
;;	       ("AtkImplementorIface" "GtkBuildable")
;;	       :type-initializer "gtk_expander_get_type")
;;   ((expanded expander-expanded "expanded" "gboolean" t t)
;;    (label expander-label "label" "gchararray" t t)
;;    (label-widget expander-label-widget "label-widget"
;;		 "GtkWidget" t t)
;;    (spacing expander-spacing "spacing" "gint" t t)
;;    (use-markup expander-use-markup "use-markup" "gboolean"
;;	       t t)
;;    (use-underline expander-use-underline "use-underline"
;;		  "gboolean" t t)))

(defmethod emit ((node interface) (target (eql :gobject-cffi)))
  (with-emitter-symbols
    (bind (((:accessors (name         node-name)
			(type-getter  class-type-getter)
			(type         class-type-name)
			(properties   class-properties)
			(fields       class-fields)) node)
	   (lisp-type  (gobject::g-name->name name))
	   (slot-specs (mapcar #'recur properties)))
    (eval-and-print
     `(gobject:define-g-interface ,type ,lisp-type
	  (:export           t
	   :type-initializer ,type-getter)
	,@slot-specs)))))

(defmethod emit ((node record) (target (eql :gobject-cffi)))
  (warn "emit is not implemented for record nodes; ignoring ~A." node))

(defmethod emit ((node property) (target (eql :gobject-cffi)))
  (bind (((:accessors (name node-name)) node)
	 (class (second (context-stack *context*)))
	 (lisp-name     (intern (string-upcase name) (current-package *context*)))
	 (accessor-name (intern (concatenate 'string
					     (string-upcase (node-name class))
					     "-" (string-upcase name))
				(current-package *context*)))
	 (type-name     (node-name (node-type node))))
    `(,lisp-name ,accessor-name ,name ,type-name t t)))

(defmethod emit ((node field) (target (eql :gobject-cffi)))
  (with-emitter-symbols
    (bind (((:accessors (name node-name)) node)
	   (class         (second stack))
	   (lisp-name     (intern (string-upcase name) package))
	   (accessor-name (intern (concatenate 'string
					       (string-upcase (node-name class))
					       "-" (string-upcase name))
				  package))
	   (type-name     (node-name (node-type node))))
      `(,lisp-name ,accessor-name ,name ,type-name t t) ;; -> gobject-property
      ;; `(:cffi ,lisp-name ,accessor-name ,name ,type-name t t) ;; -> cffi-property
      )))

(defmethod emit ((node constructor) (target (eql :gobject-cffi)))
  (warn "emit is not implemented for constructor nodes. Ignoring ~A." node))

(defmethod emit ((node method1) (target (eql :gobject-cffi)))
  (with-emitter-symbols
    (bind (((:accessors (name         node-name)
			(foreign-name node-foreign-name)
			(doc          node-doc)
			(return-type  function-return-type)
			(parameters   function-parameters)) node)
	   (parameter-names    (mapcar
				(compose #'intern #'node-name)
				parameters)) ;; TODO not good
	   ;;
	   (lisp-function-name (intern name package))
	   (lisp-method-name   (let ((gobject:*lisp-name-package*
				      package))
				 (gobject::g-name->name name))) ;; TODO make a function for this
	   ;;(return-type (xml-element-child element 'return-value))
	   ;;
	   (ffi-return-type    (gir-type->ffi-type return-type))
	   (ffi-parameters     (iter (for param in parameters)
				     (bind (((:accessors (name node-name)
							 (type parameter-type))
					     param)
					    (ffi-type (gir-type->ffi-type type)))
				       (list name ffi-type)))))

      (eval-and-print
       `(defcfun (,lisp-function-name ,foreign-name) ,ffi-return-type
	  ,@ffi-parameters
	  ))

      (eval-and-print
       `(defmethod ,lisp-method-name (,@parameter-names) ;; TODO instance parameter
	  ,@(when doc (list doc))
	  (,lisp-function-name ,@parameter-names))))))

(defmethod emit ((node virtual-method) (target (eql :gobject-cffi)))
  (with-emitter-symbols
    (bind (((:accessors (name         node-name)
			(doc          node-doc)
			(return-type  function-return-type)
			(parameters   function-parameters)) node)
	   (parameter-names    (mapcar
				(compose #'intern #'node-name)
				parameters)) ;; TODO not good
	   ;;
	   (lisp-method-name   (let ((gobject:*lisp-name-package*
				      package))
				 (gobject::g-name->name name))))

      (eval-and-print
       `(defmethod ,lisp-method-name (,@parameter-names) ;; TODO instance parameter
	  ,@(when doc (list doc))
	  (error "virtual method called"))))))

(defmethod emit ((node constant) (target (eql :gobject-cffi)))
  (bind (((:accessors (name  node-name)
		      (value node-value)) node))
	(eval
	 `(defconstant ,(intern (format nil "+~A+" name)) ;;(string-upcase name)))
	    ,(coerce (read-from-string value) 'integer)))))


;;; Type translation for ffi
;;

;;(defconstant
(defvar +fundamental-type->ffi-type+ ;; TODO name should not contain -> since there is no preferred direction
  '(("void"     . :void)
    ("char8"    . :char)
    ("int"      . :int)
    ("char*"    . :string)

    ;; TODO should the gobject be in the fundamental type list?
    ("gboolean" . :boolean)

    ("gint"     . :int)
    ("guint"    . :uint)
    ("glong"    . :long)
    ("gulong"   . :ulong)
    ("gchar"    . :char)

    ("gint16"   . :int16)
    ("guint16"  . :uint16)
    ("gint32"   . :int32)
    ("guint32"  . :uint32)

    ("gfloat"   . :float))
  "Mapping from gir fundamental type strings to cffi types.")

(defgeneric gir-type->ffi-type (type)
  (:documentation
   "Map gir type TYPE to a corresponding cffi type."))

(defmethod gir-type->ffi-type ((type type1))
  (gir-type->ffi-type (node-foreign-name type)))

(defmethod gir-type->ffi-type ((type string))
  (cond
    ;; Process pointer types recursively.
    ((ends-with #\* type)
     `(:pointer
       ,(gir-type->ffi-type (subseq type 0 (1- (length type))))))

    ;; Check fundamental type.
    ((cdr (assoc type +fundamental-type->ffi-type+
		 :test #'string=)))

    ;; Check for gobject type.


    ;; Otherwise signal error.
    (t
     (error "no type for ~S" type)))) ;; TODO condition
