;;; xml.lisp --- Read XML representation of GI repository.
;;
;; Copyright (C) 2010 Jan Moringen
;;
;; Author: Jan Moringen <jmoringe@techfak.uni-bielefeld.de>
;;
;; This Program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.
;;
;; This Program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program. If not, see <http://www.gnu.org/licenses>.

(in-package :gir.repository)


;;; Interface to Repository reading
;;

(defmethod read-repository ((source stream) (format (eql :xml)))
  "Read repository in XML-format from SOURCE."
  ;;(let ((s-xml:*auto-create-namespace-packages* nil))
  (first (process-element
	  (parse-xml source :output-type :xml-struct)
	  :repository)))

(defmethod read-repository :around ((source pathname) (format (eql :xml)))
  "Set the name slot of the repository node based on SOURCE."
  (let ((repository (call-next-method)))
    (setf (node-name repository) (pathname-name source))
    repository))


;;;
;;

#.(progn
    (unless (find-package :|ns-0|)
      (make-package :|ns-0|))
    (let ((*package* (find-package :|ns-0|)))
      (export (intern "type")))

    (unless (find-package :|c|)
      (make-package :|c|))
    (let ((*package* (find-package :|c|)))
      (export (intern "type"))
      (export (intern "prefix"))
      (export (intern "identifier")))

    (unless (find-package :|glib|)
      (make-package :|glib|))
    (let ((*package* (find-package :|glib|)))
      (export (intern "nick"))
      (export (intern "type-name"))
      (export (intern "get-type"))))

(defgeneric process-element (element type)
  (:documentation
   ""))

(defmethod process-element ((element t) (type t))
  (warn "Skipping element of unknown type ~A~%" type))

(defun traverse-children (element &key (test (constantly t)))
  "DOC"
  (iter (for child in (xml-element-children element))
	(when (and (xml-element-p child)
		   (funcall test child))
	  (collect
	      (process-element
	       child
	       (type-keyword-for-element child))))))

(defmacro def-element (type attributes &key recursive? body) ;; TODO recursive? should be composite?
  "Define a processing method for XML elements of element type TYPE
with attributes ATTRIBUTES. If RECURSIVE? is non nil, child elements
are also processed. "
  (when (member '&default-name attributes)
    (removef attributes '&default-name)
    (push '(:name            :|name|)       attributes)
    (push '(:foreign-name |c|:|identifier|) attributes)
    (push '(:doc             :|doc|)        attributes))

  (bind ((specializer (if (listp type) (first type) type))
	 (class       (if (listp type)
			  (second type)
			  (intern (symbol-name type)
				  (find-package :gir))))
	 ((:flet make-coerce (attr type))
	  (case type
	    (:boolean `(string= ,attr "1"))
	    (:integer `(parse-integer ,attr))
	    (:version `(mapcar #'parse-integer
			       (split-sequence #\. ,attr)))
	    (t        attr)))
	 ((:values bindings initializers)
	  (iter (for (slot attr . type) in attributes)
		(let ((var (gensym)))
		  (collect   `(,var ,attr) :into bindings)
		  (appending `((when ,var
				 (list ,slot ,(make-coerce var (first type)))))
			     :into initializers))
		(finally (return (values bindings initializers))))))

    `(defmethod process-element ((element t) (type (eql ,specializer)))
       (bind (((:attributes ,@bindings) element)
	      (node (apply #'make-instance
			   (quote ,class)
			   ,@(when recursive?
				   '(:children
				     (apply
				      #'append (traverse-children element))))
			   (append ,@initializers))))
	 ,@(when body
	     (list `(progn ,@body)))
         (list node)))))

(defmacro def-pass-through-element (type &key singleton?)
  `(defmethod process-element ((element t) (type (eql ,type)))
     (let ((children (apply #'append (traverse-children element))))
       ,(if singleton? 'children 'children))))


;;; Definitions
;;

;; TODO does not seem to have a name
(def-element :repository (&default-name) :recursive? t)

(def-element :include (&default-name))

(defmethod process-element ((element t) (type (eql :package)))
  "DOC"
  (bind (((:attributes (name :|name|)) element)
	 (index   (position #\- name :from-end t))
	 (version (if index
		      (read-from-string (subseq name (1+ index)))
		      1.0))
	 (name (subseq name 0 index)))))

(def-element :namespace
    ((:name              :|name|)
     (:prefix         |c|:|prefix|)
     (:shared-library    :|shared-library|))
  :recursive? t)

(def-element :alias
    ((:name      :|name|)
     (:type   |c|:|type|)
     (:target    :|target|)))

(def-element (:function function1) (&default-name) :recursive? t)

(def-element (:type type1)
    ((:name            :|name|)
     (:foreign-name |c|:|type|)
     (:doc             :|doc|)))

(def-element (:array array1)
    ((:foreign-name |c|:|type|)
     (:fixed-size      :|fixed-size| :integer)
     (:length          :|length|     :integer))
  :recursive? t)

(def-element :callback
    ((:name            :|name|)
     (:foreign-name |c|:|type|)
     (:doc             :|doc|))
  :recursive? t)

(def-element :constant
    (&default-name
     (:value :|value|)))

(def-element (:union union1)
    ((:name              :|name|)
     ;;(:doc               :|doc|)
     ;;(:version           :|version|)
     ;;(                |c|:|type|)
     (:type-name   |glib|:|type-name|)
     (:type-getter |glib|:|get-type|))
  :recursive? t)

(def-element :enumeration
    (&default-name
     (:type-name   |glib|:|type-name|) ;; also |c|:type?
     (:type-getter |glib|:|get-type|))
  :recursive? t)

(def-element (:member member1)
    (&default-name
     (:value       :|value|)
     (:nick  |glib|:|nick|)))

#|
<bitfield name="ModifierType"
doc="Masks applied to a #ClutterEvent by modifiers."
version="0.4"
glib:type-name="ClutterModifierType"
glib:get-type="clutter_modifier_type_get_type"
c:type="ClutterModifierType">
|#

(def-element :bitfield
    ((:name              :|name|)
     (:doc               :|doc|)
     (:version           :|version|   :version)
     (:type-name   |glib|:|type-name|)
     (:type-getter |glib|:|get-type|))
  :recursive? t)

(def-element (:class class1)
    (&default-name
     (:version           :|version|   :version)
     (:type-name   |glib|:|type-name|)
     (:type-getter |glib|:|get-type|)
     (:parent            :|parent|)
     (:abstract          :|abstract|  :boolean))
  :recursive? t
  :body (
	 (setf (class-implemented-interfaces node)
	       (traverse-children
		element
		:test #'(lambda (node)
			  (eq (xml-element-name node)
			      :|implements|)))))) ;; TODO get rid of this body

(defmethod process-element ((element t) (type (eql :implements)))
  nil)

(def-element :interface
    (&default-name
     (:version           :|version|  :version)
     (:type-name   |glib|:|type-name|)
     (:type-getter |glib|:|get-type|))
  :recursive? t)

(def-element :record
    (&default-name
     (:type-getter |glib|:|get-type|))
  :recursive? t)

(def-element :property
    (&default-name
     (:writable :|writable| :boolean))
  :recursive? t)

(def-element :field (&default-name) :recursive? t)

(def-element (:signal signal1)
    (&default-name
     (:version :|version| :version))
  :recursive? t)

(def-element :constructor (&default-name) :recursive? t)

(def-element :virtual-method
    (&default-name
     (:invoker :|invoker|))
  :recursive? t)

(def-element (:method method1)
    (&default-name)
  :recursive? t)

(def-pass-through-element :return-value :singleton? t)

(def-pass-through-element :parameters)

(def-element :parameter
    ((:name      :|name|)
     (:direction :|direction|))
  :recursive? t) ;; TODO doc?

(def-pass-through-element :varargs) ;; TODO (def-ignored-element)?

(defun type-keyword-for-element (element)
  "Return a the XML element name of ELEMENT as a keyword."
  (intern (string-upcase (xml-element-name element)) :keyword))
